# -*- coding: utf-8; -*-
"""add tender

Revision ID: c40a6ec00428
Revises: f8df04546a59
Create Date: 2023-09-26 18:22:17.804658

"""

# revision identifiers, used by Alembic.
revision = 'c40a6ec00428'
down_revision = 'f8df04546a59'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # corepos_tender
    op.create_table('corepos_tender',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('corepos_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['uuid'], ['tender.uuid'], name='corepos_tender_fk_tender'),
                    sa.PrimaryKeyConstraint('uuid')
                    )
    op.create_table('corepos_tender_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('corepos_id', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_corepos_tender_version_end_transaction_id'), 'corepos_tender_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_corepos_tender_version_operation_type'), 'corepos_tender_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_corepos_tender_version_transaction_id'), 'corepos_tender_version', ['transaction_id'], unique=False)


def downgrade():

    # corepos_tender
    op.drop_table('corepos_tender')
    op.drop_index(op.f('ix_corepos_tender_version_transaction_id'), table_name='corepos_tender_version')
    op.drop_index(op.f('ix_corepos_tender_version_operation_type'), table_name='corepos_tender_version')
    op.drop_index(op.f('ix_corepos_tender_version_end_transaction_id'), table_name='corepos_tender_version')
    op.drop_table('corepos_tender_version')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE POS (API) -> Rattail data importing
"""

import datetime
import decimal
import logging
from collections import OrderedDict

import sqlalchemy as sa
from sqlalchemy import orm

from rattail import importing
from rattail.db.util import normalize_full_name
from rattail_corepos import importing as corepos_importing
from rattail_corepos.corepos.util import get_core_members


log = logging.getLogger(__name__)


class FromCOREPOSToRattail(importing.ToRattailHandler):
    """
    Import handler for data coming from a CORE POS API.
    """
    host_key = 'corepos_api'
    generic_host_title = "CORE Office (API)"
    host_title = "CORE-POS (API)"
    safe_for_web_app = True

    def get_importers(self):
        importers = OrderedDict()
        importers['Customer'] = CustomerImporter
        importers['CustomerShopper'] = CustomerShopperImporter
        importers['MembershipType'] = MembershipTypeImporter
        importers['Member'] = MemberImporter
        importers['Employee'] = EmployeeImporter
        importers['Store'] = StoreImporter
        importers['Department'] = DepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Vendor'] = VendorImporter
        importers['Product'] = ProductImporter
        importers['ProductCost'] = ProductCostImporter
        importers['ProductMovement'] = ProductMovementImporter
        return importers

    def get_default_keys(self):
        keys = super().get_default_keys()

        if 'ProductMovement' in keys:
            keys.remove('ProductMovement')
        return keys

    def get_core_members(self, api, progress=None):
        if not hasattr(self, 'cached_core_members'):
            self.cached_core_members = get_core_members(
                self.config, api, progress=progress)
        return self.cached_core_members


class FromCOREPOSAPI(importing.Importer):
    """
    Base class for all CORE POS API data importers.
    """

    def setup(self):
        super().setup()

        self.establish_api()

        self.ignore_new_members = self.should_ignore_new_members()

    def datasync_setup(self):
        super().datasync_setup()

        self.establish_api()

    def establish_api(self):
        self.api = self.app.get_corepos_handler().make_webapi()

    def should_ignore_new_members(self):
        if hasattr(self, 'ignore_new_members'):
            return self.ignore_new_members

        return self.config.getbool('rattail_corepos',
                                   'importing_ignore_new_members',
                                   default=False)

    def get_core_members(self):
        members = self.handler.get_core_members(self.api, progress=self.progress)

        # maybe ignore NEW MEMBER accounts
        if self.should_ignore_new_members():
            members = [member for member in members
                       if not self.is_new_member(member)]

        return members

    def is_new_member(self, member):
        """
        Convenience method to check if the given member represents a
        "NEW MEMBER" record in CORE-POS, and hence it should be
        ignored for the import.
        """
        customers = member['customers']
        if customers:
            customer = customers[0]
            if (not customer['firstName']
                and customer['lastName'] == 'NEW MEMBER'):
                return True
        return False


class CustomerImporter(FromCOREPOSAPI, corepos_importing.model.CustomerImporter):
    """
    Importer for customer data from CORE POS API.
    """
    key = 'corepos_card_number'
    supported_fields = [
        'corepos_card_number',
        'corepos_account_id',
        'number',
        'name',
        # 'account_holder_first_name',
        # 'account_holder_last_name',
        'address_street',
        'address_street2',
        'address_city',
        'address_state',
        'address_zipcode',
    ]

    def get_host_objects(self):
        return self.get_core_members()

    def normalize_host_object(self, member):
        card_number = int(member['cardNo'])

        # figure out the "account holder" customer for the member.  note that
        # we only use this to determine the `Customer.name` in Rattail
        customers = member['customers']
        account_holders = [customer for customer in customers
                           if customer['accountHolder']]
        if account_holders:
            if len(account_holders) > 1:
                log.warning("member %s has %s account holders in CORE: %s",
                            member['cardNo'], len(account_holders), member)
            customer = account_holders[0]
        elif customers:
            if len(customers) > 1:
                log.warning("member %s has %s customers but no account holders: %s",
                            member['cardNo'], len(customers), member)
            customer = customers[0]
        else:
            raise NotImplementedError("TODO: how to handle member with no customers?")

        data = {
            'corepos_card_number': card_number,
            'corepos_account_id': int(member['customerAccountID']),
            'number': card_number,
            'name': normalize_full_name(customer['firstName'],
                                        customer['lastName']),

            # 'account_holder_first_name': customer['firstName'],
            # 'account_holder_last_name': customer['lastName'],
            'address_street': member['addressFirstLine'] or None,
            'address_street2': member['addressSecondLine'] or None,
            'address_city': member['city'] or None,
            'address_state': member['state'] or None,
            'address_zipcode': member['zip'] or None,
        }

        return data


class CustomerShopperImporter(FromCOREPOSAPI, corepos_importing.model.CustomerShopperImporter):
    """
    Importer for customer shopper data from CORE POS API.
    """
    key = ('customer_uuid', 'shopper_number')
    supported_fields = [
        'customer_uuid',
        'shopper_number',
        'corepos_customer_id',
        'first_name',
        'last_name',
        'display_name',
        'active',
        'phone_number',
        'phone_number_2',
        'email_address',
        'account_holder',
    ]

    def setup(self):
        super().setup()
        model = self.model

        self.maxlen_phone_number = self.get_maxlen_phone_number()

        self.customers_by_card_number = self.app.cache_model(
            self.session,
            model.Customer,
            key='corepos_card_number',
            query_options=[orm.joinedload(model.Customer._corepos)])

    def get_maxlen_phone_number(self):
        if hasattr(self, 'maxlen_phone_number'):
            return self.maxlen_phone_number

        model = self.model
        return self.app.maxlen(model.PhoneNumber.number)

    def get_host_objects(self):

        # first get all member data from CORE API
        members = self.get_core_members()
        normalized = []

        # then collect all the "shopper" records
        def normalize(member, i):
            normalized.extend(self.get_shoppers_for_member(member))

        self.progress_loop(normalize, members,
                           message="Collecting Person data from CORE")
        return normalized

    def get_customer_by_card_number(self, card_number):
        if hasattr(self, 'customers_by_card_number'):
            return self.customers_by_card_number.get(card_number)

        model = self.model
        try:
            return self.session.query(model.Customer)\
                               .join(model.CoreCustomer)\
                               .filter(model.CoreCustomer.corepos_card_number == card_number)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def get_shoppers_for_member(self, member):
        """
        Return a list of shopper info dicts associated with the given
        member info dict (latter having come from the CORE API).
        """
        customers = member['customers']
        shoppers = []

        # make sure account holder is listed first
        account_holder = None
        secondary = False
        mixedup = False
        for customer in customers:
            if customer['accountHolder'] and not secondary:
                account_holder = customer
            elif not customer['accountHolder']:
                secondary = True
            elif customer['accountHolder'] and secondary:
                mixedup = True
        if mixedup:
            raise NotImplementedError("TODO: should re-sort the customers list for member {}".format(member['cardNo']))

        for i, customer in enumerate(customers, 1):
            shopper = dict(customer)
            shopper['card_number'] = member['cardNo']
            shopper['shopper_number'] = i
            shoppers.append(shopper)

        return shoppers

    def normalize_host_object(self, shopper):
        card_number = shopper['card_number']

        customer = self.get_customer_by_card_number(card_number)
        if not customer:
            log.warning("Rattail customer not found for CardNo %s: %s",
                        card_number, shopper)
            return

        data = {
            'customer_uuid': customer.uuid,
            'shopper_number': shopper['shopper_number'],
            'corepos_customer_id': int(shopper['customerID']),

            'first_name': shopper['firstName'],
            'last_name': shopper['lastName'],
            'display_name': normalize_full_name(shopper['firstName'],
                                                shopper['lastName']),

            # TODO: can a CORE shopper be *not* active?
            'active': True,

            'phone_number': shopper['phone'] or None,
            'phone_number_2': shopper['altPhone'] or None,

            'email_address': shopper['email'] or None,
            'account_holder': bool(shopper['accountHolder']),
        }

        # truncate phone number data if needed
        maxlen_phone_number = self.get_maxlen_phone_number()
        if data['phone_number'] and len(data['phone_number']) > maxlen_phone_number:
            log.warning("phone_number is too long (%s chars), "
                        "will truncate to %s chars: %s",
                        len(data['phone_number']),
                        maxlen_phone_number,
                        data['phone_number'])
            data['phone_number'] = data['phone_number'][:maxlen_phone_number]
        if data['phone_number_2'] and len(data['phone_number_2']) > maxlen_phone_number:
            log.warning("phone_number_2 is too long (%s chars), "
                        "will truncate to %s chars: %s",
                        len(data['phone_number_2']),
                        maxlen_phone_number,
                        data['phone_number_2'])
            data['phone_number_2'] = data['phone_number_2'][:maxlen_phone_number]

        # swap 1st and 2nd phone numbers if only latter has value
        self.prioritize_2(data, 'phone_number')

        return data


class EmployeeImporter(FromCOREPOSAPI, corepos_importing.model.EmployeeImporter):
    """
    Importer for employee data from CORE POS API.
    """
    key = 'corepos_number'
    supported_fields = [
        'corepos_number',
        'id',
        'first_name',
        'last_name',
        'full_name',
        'status',
    ]

    def get_host_objects(self):
        return self.api.get_employees()

    def normalize_host_object(self, employee):
        return {
            'corepos_number': int(employee['emp_no']),
            'id': int(employee['emp_no']),
            'first_name': employee['FirstName'],
            'last_name': employee['LastName'],
            'full_name': normalize_full_name(employee['FirstName'], employee['LastName']),
            'status': self.enum.EMPLOYEE_STATUS_CURRENT if employee['EmpActive'] == '1' else self.enum.EMPLOYEE_STATUS_FORMER,
        }


class StoreImporter(FromCOREPOSAPI, corepos_importing.model.StoreImporter):
    """
    Importer for store data from CORE POS API.
    """
    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'id',
        'name',
    ]

    def get_host_objects(self):
        return self.api.get_stores()

    def normalize_host_object(self, store):
        return {
            'corepos_id': int(store['storeID']),
            'id': str(store['storeID']),
            'name': store['description'],
        }


class DepartmentImporter(FromCOREPOSAPI, corepos_importing.model.DepartmentImporter):
    """
    Importer for department data from CORE POS API.
    """
    key = 'corepos_number'
    supported_fields = [
        'corepos_number',
        'number',
        'name',
    ]

    def get_host_objects(self):
        return self.api.get_departments()

    def normalize_host_object(self, department):
        return {
            'corepos_number': int(department['dept_no']),
            'number': int(department['dept_no']),
            'name': department['dept_name'],
        }


class SubdepartmentImporter(FromCOREPOSAPI, corepos_importing.model.SubdepartmentImporter):
    """
    Importer for subdepartment data from CORE POS API.
    """
    key = 'corepos_number'
    supported_fields = [
        'corepos_number',
        'number',
        'name',
        'department_number',
    ]

    def get_host_objects(self):
        return self.api.get_subdepartments()

    def normalize_host_object(self, subdepartment):
        department_number = None
        if 'dept_ID' in subdepartment:
            department_number = int(subdepartment['dept_ID'])

        return {
            'corepos_number': int(subdepartment['subdept_no']),
            'number': int(subdepartment['subdept_no']),
            'name': subdepartment.get('subdept_name'),
            'department_number': department_number,
        }


class VendorImporter(FromCOREPOSAPI, corepos_importing.model.VendorImporter):
    """
    Importer for vendor data from CORE POS API.
    """
    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'id',
        'name',
        'abbreviation',
        'special_discount',
        'phone_number',
        'fax_number',
        'email_address',
    ]

    def get_host_objects(self):
        return self.api.get_vendors()

    def normalize_host_object(self, vendor):
        return {
            'corepos_id': int(vendor['vendorID']),
            'id': str(vendor['vendorID']),
            'name': vendor.get('vendorName'),
            'abbreviation': vendor.get('vendorAbbreviation') or None,
            'special_discount': decimal.Decimal(vendor['discountRate']),
            'phone_number': vendor.get('phone') or None,
            'fax_number': vendor.get('fax') or None,
            'email_address': vendor.get('email') or None,
        }


class ProductImporter(FromCOREPOSAPI, corepos_importing.model.ProductImporter):
    """
    Importer for product data from CORE POS API.
    """
    key = 'uuid'
    supported_fields = [
        'uuid',
        'corepos_id',
        'item_id',
        'upc',
        'brand_name',
        'description',
        'unit_size',
        'unit_of_measure',
        'uom_abbreviation',
        'size',
        'weighed',
        'department_number',
        'subdepartment_number',
        'regular_price_price',
        'regular_price_multiple',
        'regular_price_type',
        'food_stampable',
        # 'tax1',
        # 'tax2',
        'case_size',
    ]

    def setup(self):
        super().setup()
        model = self.model

        query = self.session.query(model.Product)\
                            .join(model.CoreProduct)\
                            .filter(model.CoreProduct.corepos_id != None)\
                            .options(orm.joinedload(model.Product._corepos))
        self.core_existing = self.app.cache_model(self.session,
                                                  model.Product,
                                                  key='corepos_id',
                                                  query=query)

        self.vendor_items_by_upc = {}

        def cache(item, i):
            if item.get('upc'):
                self.vendor_items_by_upc.setdefault(item['upc'], []).append(item)

        self.progress_loop(cache, self.api.get_vendor_items(),
                           message="Caching CORE Vendor Items")

    def get_host_objects(self):
        products = OrderedDict()

        def collect(product, i):
            if product.get('upc'):
                if product['upc'] in products:
                    log.warning("duplicate UPC encountered for '%s'; will discard: %s",
                                product['upc'], product)
                else:
                    products[product['upc']] = product

        self.progress_loop(collect, self.api.get_products(),
                           message="Fetching product info from CORE-POS")
        return list(products.values())

    def identify_product(self, corepos_product):
        model = self.app.model
        corepos_id = int(corepos_product['id'])

        if hasattr(self, 'core_existing'):
            product = self.core_existing.get(corepos_id)
            if product:
                return product

        else:
            try:
                return self.session.query(model.Product)\
                                   .join(model.CoreProduct)\
                                   .filter(model.CoreProduct.corepos_id == corepos_id)\
                                   .one()
            except orm.exc.NoResultFound:
                pass

        # at this point we'll search by `Product.item_id` instead
        return self.session.query(model.Product)\
                           .outerjoin(model.CoreProduct)\
                           .filter(model.CoreProduct.corepos_id == None)\
                           .filter(model.Product.item_id == corepos_product['upc'])\
                           .first()

    def identify_product_uuid(self, corepos_product):
        product = self.identify_product(corepos_product)
        if product:
            return product.uuid
        return self.app.make_uuid()

    def get_vendor_items(self, api_product):
        if hasattr(self, 'vendor_items_by_upc'):
            return self.vendor_items_by_upc.get(api_product['upc'])

        return self.api.get_vendor_items(upc=api_product['upc'])

    def normalize_host_object(self, product):
        model = self.model
        if 'upc' not in product:
            log.warning("CORE-POS product has no UPC: %s", product)
            return

        try:
            upc = self.app.make_gpc(product['upc'], calc_check_digit='upc')
        except (TypeError, ValueError):
            log.debug("CORE POS product has invalid UPC: %s", product['upc'])
            if len(self.key) == 1 and self.key[0] == 'upc':
                return
            upc = None

        department_number = None
        if 'department' in product:
            department_number = int(product['department']) or None

        subdepartment_number = None
        if 'subdept' in product:
            subdepartment_number = int(product['subdept']) or None

        price = None
        if product.get('normal_price') is not None:
            price = decimal.Decimal(product['normal_price'])

        data = {
            'uuid': self.identify_product_uuid(product),
            'corepos_id': int(product['id']),
            'item_id': product['upc'],
            'upc': upc,
            'brand_name': product.get('brand') or None,
            'description': product.get('description') or '',

            'department_number': department_number,
            'subdepartment_number': subdepartment_number,

            'weighed': product.get('scale') == '1',
            'food_stampable': None,
            # 'tax1': product['tax'] == '1', # TODO: is this right?
            # 'tax2': product['tax'] == '2', # TODO: is this right?

            'regular_price_price': price,
            'regular_price_multiple': 1 if price is not None else None,
            'regular_price_type': self.enum.PRICE_TYPE_REGULAR if price is not None else None,
        }

        if 'foodstamp' in product:
            data['food_stampable'] = product['foodstamp'] == '1'

        if self.fields_active(self.size_fields):
            size_info = self.normalize_size_info(product)
            data.update({
                'size': size_info['size'],
                'unit_size': size_info['unit_size'],
                'unit_of_measure': size_info['uom_code'],
                'uom_abbreviation': (size_info['uom_abbrev'] or '').strip() or None,
            })

            maxval = self.app.maxval(model.Product.unit_size)
            if data['unit_size'] and data['unit_size'] >= maxval:
                log.warning("unit_size too large (%s) for product %s, will use null instead: %s",
                            data['unit_size'], data['upc'], product)
                data['unit_size'] = None

        if 'case_size' in self.fields:
            case_size = None
            items = self.get_vendor_items(product)
            if items:
                # here we sort by `modified DESC` to get the "deterministic
                # pseudo-default" vendorItems record

                def sortkey(item):
                    modified = item.get('modified')
                    if not modified:
                        return datetime.datetime(1900, 1, 1)
                    dt = datetime.datetime.strptime(modified, '%Y-%m-%d %H:%M:%S')
                    return dt

                items = sorted(items, key=sortkey, reverse=True)
                item = items[0]
                case_size = decimal.Decimal(item['units'])

            data['case_size'] = case_size

        return data


class ProductMovementImporter(FromCOREPOSAPI, corepos_importing.model.ProductImporter):
    """
    Importer for product movement data from CORE POS API.
    """
    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'last_sold',
    ]
    allow_create = False
    allow_delete = False

    def get_host_objects(self):
        return self.api.get_products()

    def normalize_host_object(self, product):

        last_sold = None
        if 'last_sold' in product:
            last_sold = datetime.datetime.strptime(product['last_sold'], '%Y-%m-%d %H:%M:%S')
            last_sold = self.app.localtime(last_sold)
            last_sold = self.app.make_utc(last_sold)

        return {
            'corepos_id': int(product['id']),
            'last_sold': last_sold,
        }


class ProductCostImporter(FromCOREPOSAPI, corepos_importing.model.ProductCostImporter):
    """
    Importer for product cost data from CORE POS API.
    """
    key = ('corepos_vendor_id', 'corepos_sku')
    supported_fields = [
        'corepos_vendor_id',
        'corepos_sku',
        'product_uuid',
        'vendor_uuid',
        'code',
        'case_size',
        'case_cost',
        'unit_cost',
        'preferred',
    ]

    def setup(self):
        super().setup()
        model = self.app.model

        query = self.session.query(model.Vendor)\
                            .join(model.CoreVendor)\
                            .filter(model.CoreVendor.corepos_id != None)
        self.vendors = self.app.cache_model(self.session,
                                            model.Vendor,
                                            query=query,
                                            key='corepos_id')

        self.corepos_products = {}

        def cache(product, i):
            if 'upc' in product:
                self.corepos_products[product['upc']] = product

        self.progress_loop(cache, self.api.get_products(),
                           message="Caching Products from CORE-POS API")

        self.products_by_item_id = self.app.cache_model(self.session,
                                                        model.Product,
                                                        key='item_id')

    def should_warn_for_missing_vendor_id(self):
        return self.config.getbool('rattail.importing.corepos.vendor_items.warn_for_missing_vendor_id',
                                   default=True)

    def get_host_objects(self):

        # first we will cache API products by upc
        products = OrderedDict()

        def cache(product, i):
            if product.get('upc'):
                products[product['upc']] = product

        self.progress_loop(cache, self.api.get_products(),
                           message="Caching product data from CORE")

        # next we cache API vendor items, also by upc
        vendor_items = {}
        warn_for_missing_vendor_id = self.should_warn_for_missing_vendor_id()

        def cache(item, i):
            if not item.get('upc'):
                log.warning("CORE vendor item has no upc: %s", item)
                return
            if item['vendorID'] == '0':
                logger = log.warning if warn_for_missing_vendor_id else log.debug
                logger("CORE vendor item has no vendorID: %s", item)
                return
            vendor_items.setdefault(item['upc'], []).append(item)

        self.progress_loop(cache, self.api.get_vendor_items(),
                           message="Caching vendor item data from CORE")

        # now we must "sort" the vendor items for each upc.  to do
        # this we just ensure the item for default vendor is first

        def organize(upc, i):
            product = products.get(upc)
            if not product:
                return          # product not found

            vendor_id = product['default_vendor_id']
            if not vendor_id:
                return          # product has no default vendor

            items = vendor_items[upc]
            self.sort_these_vendor_items(items, vendor_id)

        self.progress_loop(organize, list(vendor_items),
                           message="Sorting items by default vendor")

        # keep the vendor item cache for reference later
        self.api_vendor_items = vendor_items

        # host objects are the API products (in original sequence)
        return list(products.values())

    def get_vendor(self, item):
        corepos_id = int(item['vendorID'])

        if hasattr(self, 'vendors'):
            return self.vendors.get(corepos_id)

        model = self.app.model
        try:
            return self.session.query(model.Vendor)\
                               .join(model.CoreVendor)\
                               .filter(model.CoreVendor.corepos_id == corepos_id)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def get_corepos_product(self, item):
        if hasattr(self, 'corepos_products'):
            return self.corepos_products.get(item['upc'])

        return self.api.get_product(item['upc'])

    def get_product(self, item):
        item_id = item.get('upc')
        if not item_id:
            return

        if hasattr(self, 'products_by_item_id'):
            return self.products_by_item_id.get(item_id)

        model = self.model
        try:
            return self.session.query(model.Product)\
                               .filter(model.Product.item_id == item_id)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def normalize_host_data(self, host_objects=None):

        # TODO: this all seems a bit hacky but works for now..
        # could even be we don't need this method?

        if host_objects is None:
            host_objects = self.get_host_objects()
        normalized = []
        self.sorted_vendor_items = {}

        def normalize(product, i):
            if not product.get('upc'):
                log.warning("product has no upc: %s", product)
                return
            items = self.sort_vendor_items(product)
            self.sorted_vendor_items[product['upc']] = items
            for item in items:
                data = self.normalize_host_object(item)
                if data:
                    normalized.append(data)

        self.progress_loop(normalize, host_objects,
                           message=f"Reading Product data from {self.host_system_title}")
        return normalized

    def normalize_host_object(self, item):
        vendor = self.get_vendor(item)
        if not vendor:
            log.warning("CORE POS vendor not found for item: %s", item)
            return

        product = self.get_product(item)
        if not product:
            # just debug logging since this is a common scenario; the
            # CORE table is for items "available from vendor" but not
            # necssarily items carried by store
            log.debug("product not found for CORE vendor item: %s", item)
            return

        core_product = self.get_corepos_product(item)
        # if not product:
        #     log.warning("CORE POS product not found for item: %s", item)
        #     return

        case_size = decimal.Decimal(item['units'])
        unit_cost = item.get('cost')
        if unit_cost is not None:
            unit_cost = decimal.Decimal(unit_cost)
        case_cost = None
        if unit_cost is not None:
            case_cost = unit_cost * case_size

        data = {
            'corepos_vendor_id': int(item['vendorID']),
            'corepos_sku': item['sku'],
            'product_uuid': product.uuid,
            'vendor_uuid': vendor.uuid,
            'code': (item['sku'] or '').strip() or None,
            'case_size': case_size,
            'case_cost': case_cost,
            'unit_cost': unit_cost,
        }

        if self.fields_active(['preference', 'preferred']):
            items = self.get_sorted_vendor_items(item)
            i = items.index(item)
            data['preference'] = i + 1
            data['preferred'] = i == 0

        return data

    def get_sorted_vendor_items(self, item):
        if hasattr(self, 'sorted_vendor_items'):
            return self.sorted_vendor_items.get(item['upc'])

        product = self.api.get_product(item['upc'])
        return self.sort_vendor_items(product)

    def sort_vendor_items(self, product):

        # TODO: this all seems a bit hacky but works for now..

        if not product.get('upc'):
            return []

        if hasattr(self, 'api_vendor_items'):
            return self.api_vendor_items.get(product['upc'], [])

        # nb. remaining logic is for real-time datasync.  here we
        # do not have a cache of vendor items so must fetch what
        # we need from API.  unfortunately we must (?) fetch *all*
        # vendor items and then filter locally
        items = [item
                 for item in self.api.get_vendor_items()
                 if item.get('upc') == product['upc']]

        vendor_id = product['default_vendor_id']
        self.sort_these_vendor_items(items, vendor_id)
        return items

    def sort_these_vendor_items(self, items, default_vendor_id):
        for item in items:
            if item['vendorID'] == default_vendor_id:
                # found the default vendor item
                i = items.index(item)
                if i != 0:
                    # it was not first; make it so
                    items.pop(i)
                    items.insert(0, item)
                break


class MembershipTypeImporter(FromCOREPOSAPI, importing.model.MembershipTypeImporter):
    """
    Imports membership type data from CORE-POS API
    """
    key = 'number'
    supported_fields = [
        'number',
        'name',
    ]

    def get_host_objects(self):
        return self.api.get_member_types()

    def normalize_host_object(self, memtype):
        return {
            'number': int(memtype['memtype']),
            'name': memtype['memDesc'],
        }


class MemberImporter(FromCOREPOSAPI, corepos_importing.model.MemberImporter):
    """
    Importer for member data from CORE POS API.
    """
    key = 'corepos_card_number'
    supported_fields = [
        'number',
        'corepos_account_id',
        'corepos_card_number',
        'customer_uuid',
        'person_uuid',
        'person_first_name',
        'person_last_name',
        'membership_type_number',
        'joined',
        'withdrew',
        'active',
    ]

    # TODO: should make this configurable
    member_status_codes = [
        'PC',
        'TERM',
    ]

    # TODO: should make this configurable
    non_member_status_codes = [
        'REG',
        'INACT',
    ]

    def setup(self):
        super().setup()
        model = self.model

        self.customers_by_number = self.app.cache_model(
            self.session,
            model.Customer,
            key='number')

        query = self.session.query(model.Person)\
                            .outerjoin(model.Customer,
                                       model.Customer.account_holder_uuid == model.Person.uuid)\
                            .outerjoin(model.CoreCustomer)\
                            .outerjoin(model.Member,
                                       model.Member.person_uuid == model.Person.uuid)\
                            .outerjoin(model.CoreMember)\
                            .filter(sa.or_(
                                model.CoreCustomer.corepos_card_number != None,
                                model.CoreMember.corepos_card_number != None))\
                            .options(orm.joinedload(model.Person.customer_accounts)\
                                     .joinedload(model.Customer._corepos))
        def card_number(person, normal):
            customer = self.app.get_customer(person)
            if customer and customer.corepos_card_number:
                return customer.corepos_card_number
            member = self.app.get_member(person)
            if member and member.corepos_card_number:
                return member.corepos_card_number
        self.people_by_card_number = self.cache_model(model.Person, query=query,
                                                      key=card_number)

        self.membership_type_number_non_member = self.get_membership_type_number_non_member()

    def get_membership_type_number_non_member(self):
        if hasattr(self, 'membership_type_number_non_member'):
            return self.membership_type_number_non_member

        return self.config.getint('corepos.membership_type.non_member')

    def should_warn_for_unknown_membership_type(self):
        return self.config.getbool('rattail.importing.corepos.warn_for_unknown_membership_type',
                                   default=True)

    def get_host_objects(self):
        return self.get_core_members()

    def get_person(self, card_number):
        if hasattr(self, 'people_by_card_number'):
            return self.people_by_card_number.get(card_number)

        model = self.model

        try:
            return self.session.query(model.Person)\
                               .join(model.Customer,
                                     model.Customer.account_holder_uuid == model.Person.uuid)\
                               .join(model.CoreCustomer)\
                               .filter(model.CoreCustomer.corepos_card_number == card_number)\
                               .one()
        except orm.exc.NoResultFound:
            pass

        try:
            return self.session.query(model.Person)\
                               .join(model.Member,
                                     model.Member.person_uuid == model.Person.uuid)\
                               .join(model.CoreMember)\
                               .filter(model.CoreMember.corepos_card_number == card_number)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def get_customer_by_number(self, number):
        if hasattr(self, 'customers_by_number'):
            return self.customers_by_number.get(number)

        model = self.model
        try:
            return self.session.query(model.Customer)\
                               .filter(model.Customer.number == number)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def get_corepos_customer(self, corepos_member):
        for customer in corepos_member['customers']:
            if customer['accountHolder']:
                return customer

    def normalize_host_object(self, member):
        card_number = member['cardNo']
        customer = self.get_customer_by_number(card_number)
        person = self.get_person(card_number)

        # TODO: at first i was *skipping* non-member status records,
        # but since CORE sort of assumes all customers are members,
        # probably not worth making the distinction here..?  it is
        # important to import the full member info from CORE, so that
        # we have it to sync back.  therefore can't afford to "skip"
        # any member records here
        memstatus = (member['memberStatus'] or '').upper() or None
        if (memstatus not in self.member_status_codes
            and memstatus not in self.non_member_status_codes):
            log.warning("unexpected status '%s' for member %s: %s",
                        member['memberStatus'], card_number, member)

        joined = None
        if member['startDate'] and member['startDate'] != '0000-00-00 00:00:00':
            joined = datetime.datetime.strptime(member['startDate'],
                                                '%Y-%m-%d %H:%M:%S')
            joined = joined.date()
            if joined == datetime.date(1900, 1, 1):
                joined = None

        withdrew = None
        if (member['endDate']
            and member['endDate'] != '0000-00-00 00:00:00'
            and member['endDate'] != '1900-01-01 00:00:00'):
            withdrew = datetime.datetime.strptime(member['endDate'],
                                                  '%Y-%m-%d %H:%M:%S')
            withdrew = withdrew.date()
            if withdrew == datetime.date(1900, 1, 1):
                withdrew = None

        typeno = int(member['customerTypeID'] or 0)
        memtype = self.get_membership_type_by_number(typeno)
        if not memtype:
            typeno = self.get_membership_type_number_non_member()
            if typeno is not None:
                memtype = self.get_membership_type_by_number(typeno)
                if not memtype:
                    raise ValueError("configured membership type for non-members is invalid!")

            logger = log.warning if self.should_warn_for_unknown_membership_type() else log.debug
            logger("unknown customerTypeID (membership_type_number) '%s' for: %s",
                   member['customerTypeID'], member)
            if typeno is not None:
                log.debug("(will override with membership_type_number: %s)", typeno)

        data = {
            'number': card_number,
            'corepos_account_id': int(member['customerAccountID']),
            'corepos_card_number': card_number,
            'customer_uuid': customer.uuid if customer else None,
            'person_uuid': person.uuid if person else None,
            'person_first_name': None,
            'person_last_name': None,
            'membership_type_number': typeno,
            'joined': joined,
            'withdrew': withdrew,
            'active': not bool(withdrew),
        }

        corepos_customer = self.get_corepos_customer(member)
        if corepos_customer:
            data['person_first_name'] = corepos_customer['firstName']
            data['person_last_name'] = corepos_customer['lastName']

        return data



# -*- coding: utf-8; -*-
"""add store extension

Revision ID: 7fea5aebddfb
Revises: 130e4632a28a
Create Date: 2021-01-27 21:01:59.767884

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '7fea5aebddfb'
down_revision = '130e4632a28a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # corepos_store
    op.create_table('corepos_store',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('corepos_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['uuid'], ['store.uuid'], name='corepos_store_fk_store'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('corepos_store_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('corepos_id', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_corepos_store_version_end_transaction_id'), 'corepos_store_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_corepos_store_version_operation_type'), 'corepos_store_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_corepos_store_version_transaction_id'), 'corepos_store_version', ['transaction_id'], unique=False)


def downgrade():

    # corepos_store
    op.drop_index(op.f('ix_corepos_store_version_transaction_id'), table_name='corepos_store_version')
    op.drop_index(op.f('ix_corepos_store_version_operation_type'), table_name='corepos_store_version')
    op.drop_index(op.f('ix_corepos_store_version_end_transaction_id'), table_name='corepos_store_version')
    op.drop_table('corepos_store_version')
    op.drop_table('corepos_store')



# -*- coding: utf-8; -*-
"""fix FK for CoreProductCost

Revision ID: b5a8734b1fe0
Revises: 15bf65f68c52
Create Date: 2023-10-20 11:20:09.231682

"""

# revision identifiers, used by Alembic.
revision = 'b5a8734b1fe0'
down_revision = '15bf65f68c52'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # corepos_product_cost
    op.alter_column('corepos_product_cost', 'corepos_id', existing_type=sa.INTEGER(), nullable=True)
    op.add_column('corepos_product_cost', sa.Column('corepos_vendor_id', sa.Integer(), nullable=False))
    op.add_column('corepos_product_cost', sa.Column('corepos_sku', sa.String(length=13), nullable=False))
    op.add_column('corepos_product_cost_version', sa.Column('corepos_vendor_id', sa.Integer(), autoincrement=False, nullable=True))
    op.add_column('corepos_product_cost_version', sa.Column('corepos_sku', sa.String(length=13), autoincrement=False, nullable=True))


def downgrade():

    # corepos_product_cost
    op.drop_column('corepos_product_cost_version', 'corepos_sku')
    op.drop_column('corepos_product_cost_version', 'corepos_vendor_id')
    op.drop_column('corepos_product_cost', 'corepos_sku')
    op.drop_column('corepos_product_cost', 'corepos_vendor_id')
    op.alter_column('corepos_product_cost', 'corepos_id', existing_type=sa.INTEGER(), nullable=False)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE Office - datasync triggers
"""

import sqlalchemy as sa

from rattail.app import GenericHandler
from rattail_corepos.datasync.corepos import make_changes_table


class CoreTriggerHandler(GenericHandler):
    """
    Handler to install and show status of CORE DB triggers, for use
    with Rattail DataSync.
    """
    supported_triggers = [
        'custdata',
        'meminfo',
        'employees',
        'departments',
        'subdepts',
        'vendors',
        'products',
        'vendorItems',
    ]

    def show_status(self, op_session, table_name):
        """
        Show trigger status for an ``office_op`` database.
        """
        print()
        print("database")
        # nb. use repr() to hide password
        print(f"url: {repr(op_session.bind.url)}")
        exists = self.database_exists(op_session)
        print(f"exists: {exists}")
        if not exists:
            return # nothing more to test
        print()

        print("changes table")
        print(f"name: {table_name}")
        table = self.make_changes_table(table_name)
        exists = self.changes_table_exists(op_session, table)
        print(f"exists: {exists}")
        if exists:
            records = op_session.execute(table.select())
            print(f"records: {len(records.fetchall())}")
        print()

        for trigger in self.supported_triggers:
            print(f"triggers for {trigger}")
            
            create = f'record_{trigger}_create'
            exists = self.trigger_exists(op_session, create)
            print(f"{create:40s} exists: {exists}")

            update = f'record_{trigger}_update'
            exists = self.trigger_exists(op_session, update)
            print(f"{update:40s} exists: {exists}")

            delete = f'record_{trigger}_delete'
            exists = self.trigger_exists(op_session, delete)
            print(f"{delete:40s} exists: {exists}")

            print()

    def database_exists(self, op_session):
        corepos = self.app.get_corepos_handler()
        op_model = corepos.get_model_office_op()
        try:
            # just query a basic table, if things are normal then we're good
            op_session.query(op_model.Department).count()
        except sa.exc.ProgrammingError:
            return False
        return True

    def trigger_exists(self, op_session, trigger):
        dbname = op_session.bind.url.database
        sql = sa.text(f"""
        SHOW TRIGGERS FROM `{dbname}` WHERE `Trigger` = :trigger
        """)
        result = op_session.execute(sql, {'trigger': trigger})
        if result.fetchone():
            return True
        return False

    def changes_table_exists(self, op_session, table):
        if isinstance(table, str):
            table = self.make_changes_table(table)
        try:
            op_session.execute(table.select())
        except sa.exc.ProgrammingError:
            return False
        return True

    def make_changes_table(self, table_name):
        metadata = sa.MetaData()
        table = make_changes_table(table_name, metadata)
        return table

    def install_all(self, op_session, table_name, dry_run=False):
        self.install_changes_table(op_session, table_name, dry_run=dry_run)
        self.install_triggers(op_session, table_name, dry_run=dry_run)

    def install_changes_table(self, op_session, table_name, dry_run=False):
        print()
        print("installing changes table...")
        print(f"{table_name}: ", end='')

        table = self.make_changes_table(table_name)
        if self.changes_table_exists(op_session, table):
            print("already exists")
            print()
            return

        if not dry_run:
            table.create(op_session.bind)
        print("done")
        print()

    def install_triggers(self, op_session, table_name, dry_run=False):
        print("installing triggers...")

        for trigger in self.supported_triggers:
            if not dry_run:
                self.drop_triggers(op_session, trigger)

                meth = getattr(self, f'create_triggers_{trigger}')
                meth(op_session, table_name)

        print("done")
        print()

    def uninstall_all(self, op_session, table_name, dry_run=False):
        self.uninstall_changes_table(op_session, table_name, dry_run=dry_run)
        self.uninstall_triggers(op_session, dry_run=dry_run)

    def uninstall_changes_table(self, op_session, table_name, dry_run=False):
        print()
        print("uninstalling changes table...")

        table = self.make_changes_table(table_name)
        if not self.changes_table_exists(op_session, table):
            print("table does not exist")
            print()
            return

        if not dry_run:
            # TODO: why does this drop() method just hang forever?
            #table.drop(op_session.bind)
            op_session.execute(sa.text(f"DROP TABLE {table_name}"))
        print("done")
        print()

    def uninstall_triggers(self, op_session, dry_run=False):
        print("uninstalling triggers...")

        for trigger in self.supported_triggers:
            if not dry_run:
                self.drop_triggers(op_session, trigger)

        print("done")
        print()

    def create_triggers_custdata(self, op_session, changes_table):

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_custdata_create
        AFTER INSERT ON custdata
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Member', CONVERT(NEW.CardNo, CHAR), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_custdata_update
        AFTER UPDATE ON custdata
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Member', CONVERT(NEW.CardNo, CHAR), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_custdata_delete
        AFTER DELETE ON custdata
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Member', CONVERT(OLD.CardNo, CHAR), 1);
        """))

    def create_triggers_meminfo(self, op_session, changes_table):

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_meminfo_create
        AFTER INSERT ON meminfo
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Member', CONVERT(NEW.card_no, CHAR), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_meminfo_update
        AFTER UPDATE ON meminfo
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Member', CONVERT(NEW.card_no, CHAR), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_meminfo_delete
        AFTER DELETE ON meminfo
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Member', CONVERT(OLD.card_no, CHAR));
        """))

    def create_triggers_employees(self, op_session, changes_table):

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_employees_create
        AFTER INSERT ON employees
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Employee', CONVERT(NEW.emp_no, CHAR), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_employees_update
        AFTER UPDATE ON employees
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Employee', CONVERT(NEW.emp_no, CHAR), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_employees_delete
        AFTER DELETE ON employees
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Employee', CONVERT(OLD.emp_no, CHAR));
        """))

    def create_triggers_departments(self, op_session, changes_table):

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_departments_create
        AFTER INSERT ON departments
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Department', CONVERT(NEW.dept_no, CHAR), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_departments_update
        AFTER UPDATE ON departments
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Department', CONVERT(NEW.dept_no, CHAR), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_departments_delete
        AFTER DELETE ON departments
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Department', CONVERT(OLD.dept_no, CHAR), 1);
        """))

    def create_triggers_subdepts(self, op_session, changes_table):

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_subdepts_create
        AFTER INSERT ON subdepts
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Subdepartment', CONVERT(NEW.subdept_no, CHAR), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_subdepts_update
        AFTER UPDATE ON subdepts
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Subdepartment', CONVERT(NEW.subdept_no, CHAR), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_subdepts_delete
        AFTER DELETE ON subdepts
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Subdepartment', CONVERT(OLD.subdept_no, CHAR), 1);
        """))

    def create_triggers_vendors(self, op_session, changes_table):

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_vendors_create
        AFTER INSERT ON vendors
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Vendor', CONVERT(NEW.vendorID, CHAR), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_vendors_update
        AFTER UPDATE ON vendors
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Vendor', CONVERT(NEW.vendorID, CHAR), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_vendors_delete
        AFTER DELETE ON vendors
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Vendor', CONVERT(OLD.vendorID, CHAR), 1);
        """))

    def create_triggers_products(self, op_session, changes_table):

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_products_create
        AFTER INSERT ON products
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Product', NEW.upc, 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_products_update
        AFTER UPDATE ON products
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Product', NEW.upc, 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_products_delete
        AFTER DELETE ON products
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('Product', OLD.upc, 1);
        """))

    def create_triggers_vendorItems(self, op_session, changes_table):

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_vendorItems_create
        AFTER INSERT ON vendorItems
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('VendorItem', CONCAT_WS('|', NEW.sku, CONVERT(NEW.vendorID, CHAR)), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_vendorItems_update
        AFTER UPDATE ON vendorItems
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('VendorItem', CONCAT_WS('|', NEW.sku, CONVERT(NEW.vendorID, CHAR)), 0);
        """))

        op_session.execute(sa.text(f"""
        CREATE TRIGGER record_vendorItems_delete
        AFTER DELETE ON vendorItems
        FOR EACH ROW INSERT INTO {changes_table} (object_type, object_key, deleted) VALUES ('VendorItem', CONCAT_WS('|', OLD.sku, CONVERT(OLD.vendorID, CHAR)), 1);
        """))

    def drop_triggers(self, op_session, trigger):

        op_session.execute(sa.text(f"""
        DROP TRIGGER IF EXISTS record_{trigger}_create;
        """))

        op_session.execute(sa.text(f"""
        DROP TRIGGER IF EXISTS record_{trigger}_update;
        """))

        op_session.execute(sa.text(f"""
        DROP TRIGGER IF EXISTS record_{trigger}_delete;
        """))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE Office commands
"""

import sys

import requests
from requests.auth import HTTPDigestAuth
import typer
from typing_extensions import Annotated

from corepos.enum import CoreDbType

from rattail.commands.typer import (make_typer, typer_eager_imports,
                                    importer_command, typer_get_runas_user,
                                    file_importer_command, file_exporter_command)
from rattail.commands.importing import ImportCommandHandler
from rattail.commands.util import rprint
from rattail_corepos.config import core_office_url
from rattail_corepos.corepos.office.util import get_fannie_config_value, get_blueline_template, make_blueline


core_office_typer = make_typer(
    name='core_office',
    help="core-office -- command line interface for CORE Office"
)


@core_office_typer.command()
def anonymize(
        ctx: typer.Context,
        dbkey: Annotated[
            str,
            typer.Option(help="Config key for CORE POS database engine to be updated.  "
                         "This key must be [corepos.db.office_op] section of your "
                         "config file.")] = 'default',
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                         "occur, but rollback (abort) the transaction at the end.")] = False,
        force: Annotated[
            bool,
            typer.Option('--force', '-f',
                         help="Do not prompt for confirmation.")] = False,
):
    """
    Make anonymous (randomize) all customer names etc.
    """
    from .anonymize import Anonymizer

    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress

    if not force:
        rprint("\n[bold yellow]**WARNING** this will modify all customer (and similar) records![/bold yellow]")
        value = input("\nreally want to do this? [yN] ")
        if not value or not config.parse_bool(value):
            sys.stderr.write("user canceled\n")
            sys.exit(1)

    try:
        import names
    except ImportError:
        sys.stderr.write("must install the `names` package first!\n\n"
                         "\tpip install names\n")
        sys.exit(2)

    try:
        import us
    except ImportError:
        sys.stderr.write("must install the `us` package first!\n\n"
                         "\tpip install us\n")
        sys.exit(2)

    anonymizer = Anonymizer(config)
    anonymizer.anonymize_all(dbkey=dbkey, dry_run=dry_run,
                             progress=progress)


@core_office_typer.command()
@file_exporter_command
def export_csv(
        ctx: typer.Context,
        **kwargs
):
    """
    Export data from CORE to CSV file(s)
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_csv.from_corepos_db_office_op.export')
    kwargs['user'] = typer_get_runas_user(ctx)
    kwargs['handler_kwargs'] = {'output_dir': kwargs['output_dir']}
    handler.run(kwargs, progress=progress)


@core_office_typer.command()
def get_config_value(
        ctx: typer.Context,
        name: Annotated[
            str,
            typer.Argument(help="Name of the config value to get.  "
                           "Prefix of `FANNIE_` is not required.")] = ...,
):
    """
    Get a value from CORE Office `fannie/config.php`
    """
    config = ctx.parent.rattail_config
    value = get_fannie_config_value(config, name)
    sys.stdout.write(f"{value}\n")


@core_office_typer.command()
@file_importer_command
def import_csv(
        ctx: typer.Context,
        corepos_dbtype: Annotated[
            CoreDbType,
            typer.Option(help="Config *type* for CORE-POS database engine to which data "
                         "should be written.  This determines which config section is "
                         "used with regard to the --corepos-dbkey arg.")] = 'office_op',
        corepos_dbkey: Annotated[
            str,
            typer.Option(help="Config key for CORE-POS database engine to which data should "
                         "be written.  This key must be defined in the config section as "
                         "determined by the --corpos-dbtype arg.")] = 'default',
        **kwargs
):
    """
    Import data from CSV to a CORE Office DB
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_corepos_db_office_op.from_csv.import')
    kwargs['user'] = typer_get_runas_user(ctx)
    kwargs['handler_kwargs'] = {
        'input_dir': kwargs['input_dir'],
        'corepos_dbtype': corepos_dbtype,
        'corepos_dbkey': corepos_dbkey,
    }
    handler.run(kwargs, progress=progress)


@core_office_typer.command()
@importer_command
def import_self(
        ctx: typer.Context,
        **kwargs
):
    """
    Import data from CORE Office ("op" DB) to "self"
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_self.from_corepos_db_office_op.import')
    kwargs['user'] = typer_get_runas_user(ctx)
    handler.run(kwargs, progress=progress)


@core_office_typer.command()
def install_triggers(
        ctx: typer.Context,
        status: Annotated[
            bool,
            typer.Option('--status',
                         help="Show current status of DB, then exit.")] = False,
        uninstall: Annotated[
            bool,
            typer.Option('--uninstall',
                         help="Uninstall table and triggers, instead of install.")] = False,
        table_name: Annotated[
            str,
            typer.Option(help="Override name of \"changes\" table if needed.")] = 'datasync_changes',
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Do not (un)install anything, but show what would have been done.")] = False,
):
    """
    Install MySQL DB triggers for use with Rattail DataSync
    """
    from rattail_corepos.corepos.office.triggers import CoreTriggerHandler

    config = ctx.parent.rattail_config
    app = config.get_app()
    corepos = app.get_corepos_handler()
    op_session = corepos.make_session_office_op()
    triggers = CoreTriggerHandler(config)

    if status:
        triggers.show_status(op_session, table_name)
    elif uninstall:
        triggers.uninstall_all(op_session, table_name, dry_run=dry_run)
    else:
        triggers.install_all(op_session, table_name, dry_run=dry_run)

    op_session.commit()
    op_session.close()


@core_office_typer.command()
def patch_customer_gaps(
        ctx: typer.Context,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Do not POST anything, but log members needing it.")] = False,
):
    """
    POST to the CORE API as needed, to patch gaps for customerID
    """
    from .patcher import CustomerGapPatcher

    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    patcher = CustomerGapPatcher(config)
    patcher.run(dry_run=dry_run, progress=progress)


@core_office_typer.command()
def ping_install(
        ctx: typer.Context,
):
    """
    Ping the /install URL in CORE Office (for DB setup)
    """
    config = ctx.parent.rattail_config
    url = core_office_url(config, require=True)
    url = f'{url}/install/'

    # TODO: hacky re-using credentials from API config..
    username = config.get('corepos.api', 'htdigest.username')
    password = config.get('corepos.api', 'htdigest.password')

    session = requests.Session()
    if username and password:
        session.auth = HTTPDigestAuth(username, password)

    response = session.get(url)
    response.raise_for_status()


# discover more commands
typer_eager_imports(core_office_typer)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for CORE member batches
"""

import csv
import logging

from corepos.db.office_op import Session as CoreSession, model as corepos

from rattail.batch import BatchHandler
from rattail.db.util import maxlen
from rattail_corepos.db.model import CoreMemberBatch, CoreMemberBatchRow


log = logging.getLogger(__name__)


class CoreMemberBatchHandler(BatchHandler):
    """
    Handler for CORE member batches.
    """
    batch_model_class = CoreMemberBatch

    def should_populate(self, batch):
        if batch.input_file:
            return True
        return False

    def setup(self, batch, progress=None):
        self.core_session = CoreSession()

    setup_populate = setup
    setup_refresh = setup

    def teardown(self, batch, progress=None):
        self.core_session.close()
        del self.core_session

    teardown_populate = teardown
    teardown_refresh = teardown

    def populate(self, batch, progress=None):
        if batch.input_file:
            return self.populate_from_file(batch, progress=progress)
        raise NotImplementedError("do not know how to populate this batch")

    def populate_from_file(self, batch, progress=None):
        """
        Populate member batch from input data file.
        """
        # TODO: how should we detect what type of input file we have?
        # this way is pretty lame but sort of works for testing
        if 'member-status' in batch.input_file:
            return self.populate_from_member_file(batch, progress=progress)

        return self.populate_from_contact_file(batch, progress=progress)

    def populate_from_contact_file(self, batch, progress=None):
        """
        Populate member batch from "contact" CSV input data file.
        """
        input_path = batch.filepath(self.config, batch.input_file)
        input_file = open(input_path, 'rt')
        reader = csv.DictReader(input_file)
        data = list(reader)
        input_file.close()

        fields = [
            'first_name',
            'last_name',
            'street',
            'city',
            'state',
            'zipcode',
            'phone',
            'email1',
        ]

        batch.set_param('fields', fields)

        maxlens = {}
        for field in fields:
            maxlens[field] = maxlen(getattr(CoreMemberBatchRow, field))

        def append(csvrow, i):
            row = self.make_row()
            row.card_number_raw = csvrow['external_id']
            if csvrow['external_id'].isdigit():
                row.card_number = int(csvrow['external_id'])
            row.first_name = csvrow['first_name']
            row.last_name = csvrow['last_name']
            row.street = csvrow['primary_address1']
            row.city = csvrow['primary_city']
            row.state = csvrow['primary_state']
            row.zipcode = csvrow['primary_zip']
            row.phone = csvrow['phone_number']
            # TODO: this seems useful, but maybe in another step?
            # row.phone = self.app.format_phone_number(csvrow['phone_number'])
            row.email1 = csvrow['email']

            for field in fields:
                if len(getattr(row, field)) > maxlens[field]:
                    log.warning("%s field is %s and will be truncated to %s "
                                "for row #%s in CSV data: %s",
                                field,
                                len(getattr(row, field)),
                                maxlens[field],
                                i + 1,
                                csvrow)
                    value = getattr(row, field)
                    setattr(row, field, value[:maxlens[field]])

            self.add_row(batch, row)

        self.progress_loop(append, data, progress,
                           message="Adding initial rows to batch")

    def populate_from_member_file(self, batch, progress=None):
        """
        Populate member batch from "member" CSV input data file.
        """
        input_path = batch.filepath(self.config, batch.input_file)
        input_file = open(input_path, 'rt')
        reader = csv.DictReader(input_file)
        data = list(reader)
        input_file.close()

        fields = [
            'first_name',
            'last_name',
            'member_type_id',
        ]

        batch.set_param('fields', fields)

        def append(csvrow, i):
            row = self.make_row()
            row.card_number = int(csvrow['signup_external_id'])
            row.first_name = csvrow['signup_first_name']
            row.last_name = csvrow['signup_last_name']
            row.member_type_id = int(csvrow['member_type'])
            self.add_row(batch, row)

        self.progress_loop(append, data, progress,
                           message="Adding initial rows to batch")

    def refresh_row(self, row):
        batch = row.batch

        # clear these first in case they are set
        row.first_name_old = None
        row.last_name_old = None
        row.street_old = None
        row.city_old = None
        row.state_old = None
        row.zipcode_old = None
        row.phone_old = None
        row.email1_old = None
        row.email2_old = None
        row.member_type_id_old = None
        row.status_text = None

        if not row.card_number:
            row.status_code = row.STATUS_MEMBER_NOT_FOUND
            row.status_text = "row has no card number"
            return

        core_member = self.core_session.get(corepos.MemberInfo, row.card_number)
        if not core_member:
            row.status_code = row.STATUS_MEMBER_NOT_FOUND
            row.status_text = "matching record not found in CORE"
            return

        core_customer = core_member.customers[0] if core_member.customers else None

        row.street_old = core_member.street
        row.city_old = core_member.city
        row.state_old = core_member.state
        row.zipcode_old = core_member.zip
        row.phone_old = core_member.phone
        row.email1_old = core_member.email
        row.email2_old = core_member.email2

        if core_customer:
            row.first_name_old = core_customer.first_name
            row.last_name_old = core_customer.last_name
            row.member_type_id_old = core_customer.member_type_id

        diffs = []
        fields = batch.get_param('fields')
        for field in fields:
            if getattr(row, field) != getattr(row, '{}_old'.format(field)):
                diffs.append(field)

        if diffs:
            row.status_code = row.STATUS_FIELDS_CHANGED
            row.status_text = ', '.join(diffs)
        else:
            row.status_code = row.STATUS_NO_CHANGE

    def describe_execution(self, batch, **kwargs):
        return ("CORE will be updated, by writing SQL directly to its DB, "
                "for each row indicating a change.  Note that this will "
                "affect one or both of the following tables:\n\n"
                "- `custdata`\n"
                "- `meminfo`")

    def execute(self, batch, progress=None, **kwargs):
        """
        Update the CORE DB with changes from the batch.
        """
        # we only want to process "update member" (changed) rows
        rows = [row for row in batch.active_rows()
                if row.status_code in (row.STATUS_FIELDS_CHANGED,)]
        if not rows:
            return True

        self.update_corepos(batch, rows, progress=progress)
        return True

    def update_corepos(self, batch, rows, progress=None):
        """
        For each of the given batch rows, this will update the CORE DB
        directly via SQL, for the fields which are specified in the
        batch params.
        """
        core_session = CoreSession()
        fields = batch.get_param('fields')

        def update(row, i):
            core_member = core_session.get(corepos.MemberInfo, row.card_number)
            if not core_member:
                log.warning("CORE member not found for row %s with card number: %s",
                            row.uuid, row.card_number)
                return

            if 'street' in fields:
                core_member.street = row.street
            if 'city' in fields:
                core_member.city = row.city
            if 'state' in fields:
                core_member.state = row.state
            if 'zipcode' in fields:
                core_member.zip = row.zipcode
            if 'phone' in fields:
                core_member.phone = row.phone
            if 'email1' in fields:
                core_member.email = row.email1

            core_customer = core_member.customers[0] if core_member.customers else None
            if core_customer:

                if 'first_name' in fields:
                    core_customer.first_name = row.first_name
                if 'last_name' in fields:
                    core_customer.last_name = row.last_name

            if 'member_type_id' in fields:
                for core_customer in core_member.customers:
                    core_customer.member_type_id = row.member_type_id

        self.progress_loop(update, rows, progress,
                           message="Updating members in CORE-POS")

        core_session.commit()
        core_session.close()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Problem reports for CORE-POS
"""

from corepos.db.office_op import Session as CoreSession, model as corepos

from rattail.problems import ProblemReport


class InvalidPersonNumbers(ProblemReport):
    """
    Looks for `custdata` records in CORE which have invalid person
    number sequence.
    """
    system_key = 'corepos'
    problem_key = 'invalid_person_numbers'
    problem_title = "Invalid person numbers"

    def find_problems(self, **kwargs):
        problems = []
        core_session = CoreSession()

        core_members = core_session.query(corepos.MemberInfo)\
                                   .order_by(corepos.MemberInfo.card_number)\
                                   .all()

        def inspect(member, i):
            for j, customer in enumerate(member.customers, 1):
                if customer.person_number != j:
                    problems.append((customer, j))

        self.progress_loop(inspect, core_members,
                           message="Looking for invalid person numbers")

        core_session.close()
        return problems


class PhoneNumbersTooLong(ProblemReport):
    """
    Looks for ``meminfo`` records in CORE which have a phone number
    value which is too long to fit into the corresponding
    ``Customers`` table column in CORE.
    """
    system_key = 'corepos'
    problem_key = 'phone_numbers_too_long'
    problem_title = "Phone numbers too long"

    def find_problems(self, **kwargs):
        problems = []
        core_session = CoreSession()

        maxlen = self.app.maxlen(corepos.Customer.phone)
        core_members = core_session.query(corepos.MemberInfo)\
                                   .order_by(corepos.MemberInfo.card_number)\
                                   .all()

        def inspect(member, i):
            if member.phone and len(member.phone) > maxlen:
                problems.append(member)

        self.progress_loop(inspect, core_members,
                           message="Looking for phone numbers too long")
        core_session.close()
        return problems
